
/* ====================================================================
 * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

#include "apr_hooks.h"
#include "apr.h"
#include "apr_compat.h"
#include "apr_lib.h"
#include "apr_strings.h"
#include "apr_buckets.h"
#include "apr_md5.h"
#include "apr_network_io.h"
#include "apr_pools.h"
#include "apr_strings.h"
#include "apr_uri.h"
#include "apr_date.h"
#include "apr_fnmatch.h"
#define APR_WANT_STRFUNC
#include "apr_want.h"


#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_script.h"

#include <stdio.h>

#include <lber.h>
#include <ldap.h>

#ifdef HAVE_SSL
#include <ldap_ssl.h>
#endif /* HAVE_SSL */

#define MOD_AUTH_LDAP_VERSION_S "2.12"
#define MOD_AUTH_LDAP_SITE_URL  "http://muquit.com/"

/*
 * $Author: Muquit $
 * $Revision: 17 $
 * $Date: 9/07/03 4:48p $
 */

/*
**  mod_auth_ldap.c
**  Apache modue to use LDAP server for HTTP basic authentication.
**
**  Limitations and Comments:
**  It is simple and clean. works with Umich LDAP SDK and Netscape LDAP SDK
**
**  Development History:
**      who                  when           why
**   ma_muquit@fccc.edu      Sep-05-1998    first cut
**
**   ma_muquit@fccc.edu      Aug-19-1999    adding AuthLDAPAuthoritative
**        config file option, so that if the uid is not in the LDAP server,
**        control will be passed to lower modules.
**
**  Take.Vos@cable.a2000.nl  Sep-20-1999    fixed multiple groups free() too 
**       early. added partials DN to require group, instead of fixed group list
**
**  muquit@muquit.com        Feb-10-2000    replaced ap_getword() and my
**      rmallws() calls with  ap_getword_white(). Replaced ap_getword() and 
**      rmallws() calls for attributes with  ap_getword_conf() as 
**      suggested by Mark OLear <mgolear@ilstu.edu>. The reason to use
**      ap_getword_conf() was: if say require ou "Foo bar" is used, 
**      ap_geword() was tokenizing the string to Foo bar, but we want to
**      consider "Foo bar" as one string.
**
**  david@giffin.org         Jan-26-2001    Added a blind bind to see
**      if the user can bind with a dn of 'uid_attr=uid, base_dn' if
**      not we just keep going with the old scheme. This allows for
**      people to authenicate through a non-anonymous directory server
**      if it's set up consistantly. Also added a little better error when 
**      no user is found during the dn search.
**
**  muquit@muquit.com       Feb-03-2001     Added code to put LDAP_USER in the
**      environment if the user could authenticate successfully. Requested
**      by Ben Brewer <ben@polygon.net>.
**
**  muquit@muquit.com       Feb-05-2001     Possible memory leak. result in
**      ldap_search_s() was not freed by calling ldap_msgfree()
**
**  muquit@muquit.com       Mar-15-2001     Added Bind_DN, Bind_Pass. Added
**      env var MOD_AUTH_LDAP_VERSION. Release 2.3.
**
**  muquit@muquit.com       Mar-18-2001     added the filter option.
**
**  muquit@muquit.com       Apr-14-2001     The uids can have spaces now.
**      example: require user foo "john doe" bar 
**
**  muquit@muquit.com       Apr-11-2002     port to Apache 2.0 API
**
**  a.schild@aarboard.ch    Aug-09-2002     Implemented deref 
**       configuration options. Serialized ldap calls, since the ldap 
**       context handles are not reentrant. Added ldap_protocol_version stuff
**
**  muquit@muquit.com       Dec-22-2002     TLS support. Debug is a 
**                                          configurable option now.
**
**  muquit@muquit.com       Dec-24-2002     SSL support. a christmas eve hack
**      Coded according to sslnoauth.c comes with iplanet ldap C sdk example
**      code. Recquires iplanet C SDK and netscape navigator cert7.db with 
**      the CA certificate in it. 11:24 PM EST
**
** muquit@muquit.com        Feb-23-2003     Nested group support
** muquit@muquit.com        Mar-02-2003     new env var LDAP_USER_ALT
** I won't put comments here anymore.. May-26-2003 
**
**
* Tue Aug 26 2003 Scott Lampert <scott@lampert.org>
*
* - Added AuthOnBind apache config option
* - Added LDAP_Persistent apache config option
* - Added Bind_Tries apache config option
*
* $Id: mod_auth_ldap.c,v 1.3 2003/08/26 20:23:40 scott Exp $
*
*/

/* non ssl port */
#ifndef LDAP_PORT
#define LDAP_PORT   389
#endif /* LDAP_PORT */

#define MML    __LINE__

typedef struct _Conns
{
    LDAP
        *ld;
}Conns;

typedef struct _ldap_auth_config_rec
{
    LDAP
        *ld; /* The ldap connection. Must be protected by a semaphore*/

#if APR_HAS_THREADS
    apr_thread_mutex_t
        *mutex; /* Used to serialize access to ld*/
#endif

    char
        *ldap_server,
        *base_dn,
        *sub_dn,
        *uid_attr,
        *uid_attr_alt,
        *group_attr;

    char
        filter[512];

    char
        *user_dn;

    char
        *bind_dn,
        *bind_pass,
        *auth_on_bind;

    int
        ldap_port,
        ldap_protocol_version, /* The ldap protocoll version to use*/
        ldap_connect_timeout,
        ldap_persistent,
        max_bind_tries;

    int
        env_vars_set,
        debug,
        auth_ldapauthoritative,
        support_nested_groups;

#ifdef HAVE_LDAP_START_TLS_S
    int
        auth_ldapstarttls;
#endif /* HAVE_LDAP_START_TLS_S */

#ifdef HAVE_SSL
    char
        *ca_cert_db_path; /* path of cert7.db in format used by netscape communicator*/
#endif /* HAVE_SSL */
    int
        start_ssl;


    int
        ldap_opt_deref;

    char
        *dummy;
} ldap_auth_config_rec;

/*
** Tired of cluttered log message with defines everywhere. A function call will
** be made even if DEBUG_LDAP is not defined. But I also don't have time
** to maintain two dlls for windows either. So debug is a configurable
** option now.
*/
static void m_rerr(request_rec *r,int debug,char *format,...)
{
    va_list
        args;

    char
        buf[2048];

    if (debug)
    {
        va_start(args,format);
        /*
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,format,args);
        */
        ap_vsnprintf(buf,sizeof(buf),format,args);
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,"%s",buf);
        va_end(args);
    }
}

static void m_err(cmd_parms *cmd,int debug,char *format,...)
{
    va_list
        args;
    char
        buf[2048];

    if (debug)
    {
        va_start(args,format);
        ap_vsnprintf(buf,sizeof(buf),format,args);
        ap_log_error(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,cmd->server,
                "%s",buf);
        va_end(args);
    }
}

static void *create_ldap_auth_dir_config(apr_pool_t *p,char *d)
{
    ldap_auth_config_rec
        *cr;

    cr=(ldap_auth_config_rec *) ap_pcalloc(p,sizeof(ldap_auth_config_rec));
    if (cr == (ldap_auth_config_rec *) NULL)
        return (NULL);

    memset(cr,0,sizeof(ldap_auth_config_rec));

    cr->ld=(LDAP *) NULL;
#if APR_HAS_THREADS
    apr_thread_mutex_create(&cr->mutex, APR_THREAD_MUTEX_DEFAULT, p);
#endif /* APR_HAS_THREADS */

    cr->ldap_server=NULL; 
    cr->base_dn=NULL;
    cr->sub_dn=NULL;
    cr->user_dn=NULL;
    cr->bind_dn=NULL;
    cr->bind_pass=NULL;
    cr->max_bind_tries=10;
    cr->auth_on_bind=0;
    cr->uid_attr=ap_pstrdup(p,"uid");

    /*
     * Note about uid_attr_alt. If this attribute is specified, after
     * authenticating the use, get the value of that attribute and set it
     * to AUTH_USER env variable. Someone asked for it, because they use
     * that env var to match the ldap use to a user in some relational
     * database. Feb-23-2003 Feb-23-2003 
     */
    cr->uid_attr_alt=NULL;

    /*
     * Netscape/iPlanet LDAP server uses uniqueMember as the attribute to
     * store the user dn if the user is member of that group
     */
    cr->group_attr=ap_pstrdup(p,"uniqueMember");
    cr->ldap_port=LDAP_PORT;
    cr->ldap_protocol_version= LDAP_VERSION2;
    cr->ldap_connect_timeout=4; /* 4 seconds */
    cr->ldap_persistent=1;
    cr->env_vars_set=0;
    cr->debug=0;
    cr->auth_ldapauthoritative=1; /* fortress is secure by default */
    cr->support_nested_groups=0;

#ifdef HAVE_LDAP_START_TLS_S
    cr->auth_ldapstarttls=0;
#endif /* HAVE_LDAP_START_TLS_S */

#ifdef HAVE_SSL
    cr->ca_cert_db_path=NULL;
#endif /* HAVE_SSL */
    cr->start_ssl=0;
    cr->ldap_opt_deref= LDAP_DEREF_NEVER;  /* Default is to never deref */

    return (cr);
}

static const char *set_ldap_server(cmd_parms *cmd,void *dir_config,
                                   const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->ldap_server=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_ldap_port(cmd_parms *cmd,void *dir_config,
                                 const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->ldap_port=atoi(arg);
    return (NULL);
}

static const char *set_ldap_protocol_version(cmd_parms *cmd,void *dir_config,
        const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->ldap_protocol_version=atoi(arg);
    return (NULL);
}

static const char *set_ldap_connect_timeout(cmd_parms *cmd,void *dir_config,
        const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->ldap_connect_timeout=atoi(arg);
    return (NULL);
}

/*
static const char *set_ldap_deref(cmd_parms *cmd,ldap_auth_config_rec *cr,
                                  const char *arg)
                                  */
static const char *set_ldap_deref(cmd_parms *cmd,void *dir_config,
                                  const char *arg)
{
    ldap_auth_config_rec
        *cr=(ldap_auth_config_rec *) dir_config;

    char
        *a=ap_pstrdup(cmd->pool,arg);

    if (a != NULL)
    {
        if ((strcasecmp(a,"NEVER") == 0))
        {
            m_err(cmd,cr->debug,"[mod_auth_ldap.c] - Deref is NEVER");
            cr->ldap_opt_deref= LDAP_DEREF_NEVER;
        }
        else if ((strcasecmp(a,"SEARCHING") == 0))
        {
            m_err(cmd,cr->debug,"[mod_auth_ldap.c] - Deref is SEARCHING");
            cr->ldap_opt_deref= LDAP_DEREF_SEARCHING;
        }
        else if ((strcasecmp(a,"FINDING") == 0))
        {
            m_err(cmd,cr->debug,"[mod_auth_ldap.c] - Deref is FINDING");
            cr->ldap_opt_deref= LDAP_DEREF_FINDING;
        }
        else if ((strcasecmp(a,"ALWAYS") == 0))
        {
           m_err(cmd,cr->debug,"[mod_auth_ldap.c] - Deref is ALWAYS");
           cr->ldap_opt_deref= LDAP_DEREF_ALWAYS;
        }
        else
        {
            m_err(cmd,cr->debug,
                    "[mod_auth_ldap.c] -  Deref has unexpected argument");
        }
    }
    else
    {
        m_err(cmd,cr->debug,
                "[mod_auth_ldap.c] - Deref is missing its argument");
    }
    return (NULL);
}

static const char *set_base_dn(cmd_parms *cmd,void *dir_config,
                               const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->base_dn=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_sub_dn(cmd_parms *cmd,void *dir_config,
                               const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->sub_dn=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_bind_dn(cmd_parms *cmd,void *dir_config,
                               const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->bind_dn=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_bind_pass(cmd_parms *cmd,void *dir_config,
                                 const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->bind_pass=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_max_bind_tries(cmd_parms *cmd,void *dir_config,
        const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->max_bind_tries=atoi(arg);

    if (cr->max_bind_tries <= 0) cr->max_bind_tries=1;
    return (NULL);
}

#ifdef HAVE_SSL
static const char *set_ca_cert_db_path(cmd_parms *cmd,void *dir_config,
                                 const char *arg)
{
    ldap_auth_config_rec
        *cr;

    cr=(ldap_auth_config_rec *) dir_config;
    cr->ca_cert_db_path=ap_pstrdup(cmd->pool,arg);
    cr->start_ssl=1;
    return (NULL);
}
#endif /* HAVE_SSL */

static const char *set_uid_attr(cmd_parms *cmd,void *dir_config,
                                const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->uid_attr=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_uid_attr_alt(cmd_parms *cmd,void *dir_config,
                                const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->uid_attr_alt=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_group_attr(cmd_parms *cmd,void *dir_config,
                                const char *arg)
{
    ldap_auth_config_rec *cr=(ldap_auth_config_rec *) dir_config;
    cr->group_attr=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_ldapauthoritative(cmd_parms *cmd,ldap_auth_config_rec
        *cr,char *arg)
{
    char
        *a=ap_pstrdup(cmd->pool,arg);

    if (a != NULL)
    {
        if ((strcasecmp(a,"no") == 0) || (strcasecmp(a,"off") == 0))
        {
            m_err(cmd,cr->debug,
                    "[mod_auth_ldap.c] -  AuthLDAPAuthoritative is off");
            cr->auth_ldapauthoritative=0;
        }
        else
        {
            if ((strcasecmp(a,"yes") == 0) || (strcasecmp(a,"on") == 0))
            {
                m_err(cmd,cr->debug,
                        "[mod_auth_ldap.c] -  AuthLDAPAuthoritative is on");
                cr->auth_ldapauthoritative=1;
            }
            else
            {
                m_err(cmd,cr->debug,"[mod_auth_ldap.c] - AuthLDAPAuthoritative has unexpected argument");
            }
        }
    }
    else
    {
        m_err(cmd,cr->debug,"[mod_auth_ldap.c] - AuthLDAPAuthoritative is missing its argument");
    }
    return (NULL);
}

static const command_rec ldap_auth_cmds[]=
    {
        AP_INIT_FLAG("LDAP_Debug",ap_set_flag_slot,
                     (void *) APR_XtOffsetOf(ldap_auth_config_rec,debug),
                     OR_AUTHCFG,
                     "Print debug messages in server error log or not"),

        AP_INIT_TAKE1("LDAP_Server",set_ldap_server,
                      NULL,
                      OR_AUTHCFG,
                      "The name or IP address of the LDAP server"),

        AP_INIT_TAKE1("LDAP_Port",set_ldap_port,
                      NULL,
                      OR_AUTHCFG,
                      "The port ldap server is listening"),

        AP_INIT_TAKE1("LDAP_Protocol_Version",set_ldap_protocol_version,
                      NULL,
                      OR_AUTHCFG,
                      "The ldap protocol to use for communication with the ldap server"),

        AP_INIT_TAKE1("LDAP_Connect_Timeout",set_ldap_connect_timeout,
                      NULL,
                      OR_AUTHCFG,
                      "Timeout if could not connect in specified seconds"),


        AP_INIT_TAKE1("LDAP_Deref",set_ldap_deref,
                      NULL,
                      OR_AUTHCFG,
                      "The ldap dereferencing of alias"),

        AP_INIT_TAKE1("Base_DN",set_base_dn,
                      NULL,
                      OR_AUTHCFG,
                      "LDAP Base DN"),

        AP_INIT_TAKE1("Sub_DN",set_sub_dn,
                      NULL,
                      OR_AUTHCFG,
                      "Specify a node in the Base DN with this directive to construct a user DN for explicit binding if AuthOnBind is On"),

        AP_INIT_TAKE1("UID_Attr",set_uid_attr,
                      NULL,
                      OR_AUTHCFG,
                      "LAP attribute for uid"),


        AP_INIT_TAKE1("UID_Attr_Alt",set_uid_attr_alt,
                      NULL,
                      OR_AUTHCFG,
                      "LAP attribute to search and set to LDAP_USER_ALT after authentication"),


        AP_INIT_TAKE1("Group_Attr",set_group_attr,
                      NULL,
                      OR_AUTHCFG,
                      "Group attribute for user DN. iPlanet uses uniqueMember as the attribute"),

        AP_INIT_TAKE1("Bind_DN",set_bind_dn,
                      NULL,
                      OR_AUTHCFG,
                      "LDAP Bind DN"),

        AP_INIT_TAKE1("Bind_Pass",
                      set_bind_pass,NULL,
                      OR_AUTHCFG,
                      "LDAP Bind DN password"),

        AP_INIT_TAKE1("Bind_Tries",
	              set_max_bind_tries,NULL,
		      OR_AUTHCFG,
		      "Maximum number of times to try to bind to the LDAP server"),

        AP_INIT_FLAG("AuthOnBind",ap_set_flag_slot,
                     (void *) APR_XtOffsetOf(ldap_auth_config_rec,auth_on_bind),OR_AUTHCFG,
                     "Set to On to consider a successful bind with the authenticating user's username + Base_DN and password to be a successful authentication."),

        AP_INIT_FLAG("LDAP_Persistent",ap_set_flag_slot,
                     (void *) APR_XtOffsetOf(ldap_auth_config_rec,ldap_persistent),OR_AUTHCFG,
                     "Set to Off to make a new connection for every LDAP request."),

#ifdef HAVE_LDAP_START_TLS_S
        AP_INIT_FLAG("LDAP_StartTLS",ap_set_flag_slot,
                     (void *) APR_XtOffsetOf(ldap_auth_config_rec,auth_ldapstarttls),OR_AUTHCFG, "Set to yes to allow LDAP to enable Transport Layer Security"),
#endif /* HAVE_LDAP_START_TLS_S */

#ifdef HAVE_SSL
        AP_INIT_TAKE1("LDAP_CertDbDir",
                set_ca_cert_db_path,NULL,
                OR_AUTHCFG,
                "The full path of the directory containing certificates of Certificate Authorities in cert7.db and key3.db file used by Netscape Communicator"),
#endif /* HAVE_SSL */


        AP_INIT_FLAG("AuthLDAPAuthoritative",ap_set_flag_slot,
                     (void *) APR_XtOffsetOf(ldap_auth_config_rec,auth_ldapauthoritative),OR_AUTHCFG,
                     "Set to no to allow control to be passed along to lower modules if the uid s not known to this module"),

        AP_INIT_FLAG("SupportNestedGroups",ap_set_flag_slot,
                     (void *) APR_XtOffsetOf(ldap_auth_config_rec,support_nested_groups),OR_AUTHCFG,
                     "Set to yes if nested group support is needed. It's a expensive operation."),

        {NULL}
    };

module AP_MODULE_DECLARE_DATA auth_ldap_module;

static void mod_auth_ldap_free_dn(char *dn)
{
#if LDAP_SDK_VERSION == 2
        (void) free(dn);
#else
        ldap_memfree(dn);
#endif /* LDAP_SDK_VERSION != 2 */
}



/*
**  mod_auth_ldap_find_user_dn()
**    return the DN of the user 
**
**  Parameters:
**      r           pointer to request_rec
**      try_again   OUT. if try_again is one, then the caller can retry to bind
**                       otherwise it means the connection is ok but the user may
**                       not exist.
**
**  Return Values:
**
**      pointer to DN of the userid if found, NULL otherwise
**
**  Limitations and Comments:
**      ld must be valid.
**
**      any LDAP sdk should work. returns pointer to a malloc'd space, the 
**      caller is responsible to free it.
**
**  Development History:
**      who                  when           why
**      ma_muquit@fccc.edu   Nov-29-1998    first cut for my own web server
**                                          mhttpd
**      muquit@muquit.com    Mar-15-2001    bind with bind and pass if provided.
*/
static char *mod_auth_ldap_find_user_dn(request_rec *r,int *try_again)
{
    int
    count = 0;

    int
        bound=0,
        rc=(-1);

    LDAPMessage
        *result,
        *entry;

    char
        *dn,
        szfilter[256];

    ldap_auth_config_rec
        *cr;

    char
        *base_dn,
        *uid_attrib,
        *userid,
        *bind_dn,
        *bind_pass;

    result=(LDAPMessage *) NULL;
    entry=(LDAPMessage *) NULL;
    dn=(char *) NULL;
    *try_again=1;

    cr=(ldap_auth_config_rec *)
       ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    if (cr == NULL || cr->ld == NULL)
        return(NULL);

    base_dn=cr->base_dn;
    uid_attrib=cr->uid_attr;
    userid=r->user;
    bind_dn=cr->bind_dn;
    bind_pass=cr->bind_pass;

    /* prepare filter with UidAttr. */
    ap_snprintf(szfilter,sizeof(szfilter)-1,"(%s=%s)",uid_attrib,userid);

    m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d) ] - Using LDAP filter: %s",MML,szfilter);

    /*
    ** if the bind DN and bind password is specified, try to bind with
    ** them first.
    */
    if (bind_dn != NULL && bind_pass != NULL)
    {
        m_rerr(r,cr->debug,"[mod_auth_ldap.c] - trying to bind with bind DN \"%s\" and password (not shown)",bind_dn);

        if (ldap_simple_bind_s(cr->ld,bind_dn,bind_pass) == LDAP_SUCCESS)
        {
            bound=1;
            m_rerr(r,cr->debug,"[mod_auth_ldap.c] - Bound successfully with DN \"%s\" and password (not shown)",bind_dn);
        }
        else
        {
            m_rerr(r,1,"[mod_auth_ldap.c] - Warning: bind with DN \"%s\" and password (not shown) did not succeed",bind_dn);
        }
    }

    /* Bind anonymously now only if not bound yet. */
    if (bound == 0)
    {
        /*
        rc=ldap_simple_bind_s(cr->ld,cr->base_dn,NULL);
        */
        rc=ldap_simple_bind_s(cr->ld,NULL,NULL);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,cr->debug,
                "[mod_auth_ldap.c (%d) ] - Could not bind to LDAP server, reason: %s",MML,ldap_err2string(rc));
            return ((char *) NULL);
        }
    }

    /* search to find the dn for the user */
    rc=ldap_search_s(cr->ld,
                     base_dn,
                     LDAP_SCOPE_SUBTREE,
                     szfilter,
                     NULL,
                     0,
                     &result);
    if (rc != LDAP_SUCCESS)
    {
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                      "[mod_auth_ldap.c] - ldap_search_s() failed");

        /* note: ldap_err2string() returns pointer to a static space */
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                      "[mod_auth_ldap.c] - Error: %s",ldap_err2string(rc));
        return (NULL);
    }

    /* found something, better be a single match */
    if ((count = ldap_count_entries(cr->ld,result)) != 1)
    {
        if (count > 0)
        {
            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                  "[mod_auth_ldap.c] - multiple matches found for: %s",userid);
        }
        else
        {
            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                  "[mod_auth_ldap.c (%d) ] - no user found for: %s",MML,userid);
        }

        /* don't try again, connection is open but user not found */
         *try_again=0;

        if (result != (LDAPMessage *) NULL)
        {
            ldap_msgfree(result);
            result=NULL;
        }

        return (NULL);
    }
    if ((entry =ldap_first_entry(cr->ld,result)) == NULL)
    {
        m_rerr(r,cr->debug,
         "[mod_auth_ldap.c] - %s=%s Unknown in LDAP server",uid_attrib,userid);

        if (result != (LDAPMessage *) NULL)
        {
            ldap_msgfree(result);
            result=NULL;
        }

        /* don't try again, connection is open but user not found */
        *try_again=0;
        return (NULL);
    }

    if ((dn=ldap_get_dn(cr->ld,entry)) == NULL)
    {
        m_rerr(r,1,"[mod_auth_ldap.c] - could not obtain User DN from LDAP for user: %s",
                userid);
        return (NULL);
    }

    if (result != (LDAPMessage *) NULL) /* XXX */
    {
        ldap_msgfree(result);
        result=NULL;
    }

    return (dn); /* caller must free it */
}


/*
** remove all whitespace (leading or trailing) from a string.
** borrowed from public domain c snippets library.
** muquit, sep-07-1998
*/
static char *rmallws(char *str)
{
    char
    *obuf,
    *nbuf;

    if (str)
    {
        for (obuf=str, nbuf=str; *obuf; ++obuf)
        {
            if (!isspace(*obuf))
                *nbuf++ = *obuf;
        }
        *nbuf='\0';
    }
    return (str);
}

/* for my use only - muquit@muquit.com */
#ifdef MY_DEBUG
static void dumpResult(LDAP *ld,LDAPMessage *res,request_rec *r)
{
    LDAPMessage
    *ent;

    BerElement
    *ber;

    int
    i;

    char
    *attr,
    **vals;

    ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                  "XXXXXXXX");
    for (ent=ldap_first_entry
             (ld,res); ent != NULL; ent=ldap_next_entry
                                        (ld,ent))
    {
        for (attr=ldap_first_attribute(ld,ent,&ber); attr != NULL;
                attr=ldap_next_attribute(ld,ent,ber))
        {
            if (strcasecmp(attr,"jpegphoto") != 0)
            {
                if ((vals=ldap_get_values(ld,ent,attr)) != NULL)
                {
                    for (i=0; vals[i] != NULL; i++)
                    {
                        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                                      "dump: %s=%s",attr,vals[i]);
                    }
                    if (vals)
                        ldap_value_free(vals);
                }
            }
        }
    }
}
#endif /* MY_DEBUG */

/*
 * search cr->uid_attr_alt and set the value to LDAP_USE_ALT env var
 */
static void mod_auth_ldap_set_ldap_user_alt(request_rec *r)
{
    ldap_auth_config_rec
        *cr;

    int
        rc;

    char
        **val=NULL;

    LDAPMessage
        *res=NULL,
        *ent=NULL;


    cr= (ldap_auth_config_rec *)
        ap_get_module_config(r->per_dir_config,&auth_ldap_module);
    if (cr == NULL)
        return;

    if (cr->ld == NULL)
        return;

    (void) ap_snprintf(cr->filter,sizeof(cr->filter)-1,"(%s=*)",
                       cr->uid_attr_alt);
    rc=ldap_search_s(cr->ld,
            cr->user_dn,
            LDAP_SCOPE_SUBTREE,
            cr->filter,
            NULL,
            0,
            &res);

    if (rc != LDAP_SUCCESS)
    {
        m_rerr(r,cr->debug,
                "[mod_auth_ldap.c] %d - search failed with filter: %s",
                MML,cr->filter);
        return;
    }

    /* we;ll set the value of the first match */
    ent=ldap_first_entry(cr->ld,res);
    if (ent)
    {
        val=ldap_get_values(cr->ld,ent,cr->uid_attr_alt);
        if (val != NULL)
        {
            m_rerr(r,cr->debug,
                "[mod_auth_ldap.c] %d - %s=%s",
                MML,cr->uid_attr_alt,val[0]);
            ap_table_setn(r->subprocess_env,"LDAP_USER_ALT",val[0]);
        }
    }


    if (res != NULL)
        ldap_msgfree(res);
}

/*
** validate the user id via LDAP
** Returns:
**      DECLINED        if ldap server is not defined in config file
**      HTTP_UNAUTHORIZED   for any other kind of errors
**      OK              if authorized
*/
static int check_user_authentication(request_rec *r)
{
    ldap_auth_config_rec
        *cr;

    int
        rcode=OK,
        rc,
        ldap_vers;

    int
        timeout=0,
        ssl_ok=0,
        res;

    conn_rec
        *c;

    const char
        *sent_pw=NULL;

    char
        *dn=NULL;

    Conns
        *conns=NULL;

    int
        try_again=1,
        bind_try=1;

    char
      authonbind_dn[256];

    ssl_ok=0;
    c=r->connection;
    cr=(ldap_auth_config_rec *)
       ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    res=ap_get_basic_auth_pw(r,&sent_pw);
    if (res)
        return (res);

    /* if no user is specified, keep asking */
    if (r->user == NULL || *r->user == '\0')
    {
        m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - This LDAP auth module does not accept empty user",MML);
        if (cr->auth_ldapauthoritative)
        {
            ap_note_basic_auth_failure(r);
            return (HTTP_UNAUTHORIZED);
        }
        else
            return(DECLINED);
    }

    /* if no password is specified, keep asking */
    if ((sent_pw == NULL) || (*sent_pw == '\0'))
    {
        m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - This LDAP auth module does not accept empty password",MML);
        if (cr->auth_ldapauthoritative)
        {
            ap_note_basic_auth_failure(r);
            return (HTTP_UNAUTHORIZED);
        }
        else
            return(DECLINED);
    }

#if APR_HAS_THREADS
    if (cr->mutex)
    {
        apr_thread_mutex_lock(cr->mutex);
    }
#endif /* APR_HAS_THREADS */



#ifdef HAVE_SSL
    m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - mod_auth_ldap v%s (compiled with SSL) url: %s",MML,MOD_AUTH_LDAP_VERSION_S,MOD_AUTH_LDAP_SITE_URL);
#endif /* HAVE_SSL */

#ifdef HAVE_LDAP_START_TLS_S
    m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d) ]- mod_auth_ldap v%s (compiled with OpenLDAP TLS) url: %s",MML,MOD_AUTH_LDAP_VERSION_S,MOD_AUTH_LDAP_SITE_URL);
#endif /* HAVE_LDAP_START_TLS_S */

    if (cr->ldap_server == NULL)
    {
        m_rerr(r,1,"[mod_auth_ldap.c (%d)] - No LDAP server specified",MML);
        rcode=DECLINED;
        goto exit_processing;
    }

    if (cr->ldap_server)
    {
        m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - LDAP server=%s,Port=%d",
            MML,cr->ldap_server,cr->ldap_port);
    }


    if (! cr->start_ssl) /* we'll set ldapv3 if ssl is on */
    {
        /* Set LDAP version to whatever configured */
        ldap_set_option( NULL, LDAP_OPT_PROTOCOL_VERSION,
                &cr->ldap_protocol_version);

        /*  Set alias deference option */
        if (ldap_set_option(NULL,LDAP_OPT_DEREF,&cr->ldap_opt_deref) 
                != LDAP_SUCCESS)
        {
            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                          "[mod_auth_ldap.c] - failed setting LDAP_DEREF");
        }
    }

    /*
    ** If a previsous connection is made from that pid or thread, we'll 
    ** re-use it
    */
    conns=(Conns *) ap_get_module_config(r->server->module_config,
            &auth_ldap_module);
    if (conns == NULL)
    {
        /* we're sol'd */
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,0,r,
                "[mod_auth_ldap.c (%d)] - Could not get connections",MML);
        rcode=DECLINED;
        goto exit_processing;
    }

    bind_try=1;
try_again:

#ifdef HAVE_SSL
    if (cr->start_ssl && cr->ca_cert_db_path) /* cert7.db path specified */
    {
        /* initialize the SSL client */
        if (ldapssl_client_init(cr->ca_cert_db_path,NULL) != 0)
        {
            m_rerr(r,1,"[mod_auth_ldap.c (%d)] - Could not initialize SSL client, cert db path: %s",MML,cr->ca_cert_db_path);
            if (bind_try >= cr->max_bind_tries)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }
        }
        m_rerr(r,cr->debug,
                "[mod_auth_ldap.c (%d)] - SSL client initialized, certificate db dir %s is valid",MML,cr->ca_cert_db_path);

        /* get a handle to LDAP connection */
        if (conns->ld == NULL)
        {
                m_rerr(r,cr->debug,
                    "[mod_auth_ldap.c (%d)] - SSL MAKING NEW CONNECTION, attempt# %d, pid=%d",MML,bind_try,getpid());
                conns->ld=ldapssl_init(cr->ldap_server,cr->ldap_port,1);
                cr->ld=conns->ld;
        }
        else
        {
            cr->ld=conns->ld;
            m_rerr(r,cr->debug,
                "[mod_auth_ldap.c (%d)] - SSL using already open ldap connection: 0x%x,pid=%d",MML,cr->ld,getpid());
        }

        if (cr->ld == NULL)
        {
            m_rerr(r,1,"[mod_auth_ldap.c (%d) ] - Could not initialize SSL connection, LDAP server %s at port %d",MML,cr->ldap_server,cr->ldap_port);
            if (bind_try >= cr->max_bind_tries)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }
        }
   
        m_rerr(r,cr->debug,
                "[mod_auth_ldap.c (%d)] - Initialized SSL connection to LDAP server %s:%d with SSL",MML,cr->ldap_server,cr->ldap_port);


        /* use LDAPv3 */
        ldap_vers=LDAP_VERSION3;
        rc=ldap_set_option(cr->ld,LDAP_OPT_PROTOCOL_VERSION,&ldap_vers);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,1,"[mod_auth_ldap.c] - SSL ldap_set_option(LDAP_VERSION3) failed (%d): %s",rc,ldap_err2string(rc));
            if (bind_try >= cr->max_bind_tries)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }
        }
        ssl_ok=1;
    }
#endif /* HAVE_SSL */

    /* connect to LDAP server if we already didn't make ssl connection*/
    if (ssl_ok == 0)
    {
        /*
        if (cr->ld == NULL)
        */
        if (conns->ld == NULL)
        {
            m_rerr(r,cr->debug,
               "[mod_auth_ldap.c] (%d) - MAKING NEW CONNECTION, try# %d, pid=%d",MML,bind_try,getpid());
                conns->ld=ldap_init(cr->ldap_server,cr->ldap_port);
                cr->ld=conns->ld;

            m_rerr(r,cr->debug,
                "[mod_auth_ldap.c] (%d) - cr->ld: 0x%x, pid=%d",MML,cr->ld,getpid());
        }
        else
        {
            cr->ld=conns->ld;
            m_rerr(r,cr->debug,
                "[mod_auth_ldap.c (%d)] - Using already open ldap connection: %x, pid=%d",MML,cr->ld,getpid());
        }

        if (cr->ld == (LDAP *) NULL)
        {
            m_rerr(r,1,"[mod_auth_ldap.c (%d)] - Could not initialize connection to LDAP server %s at port %d",MML,cr->ldap_server,cr->ldap_port);
            if (bind_try >= cr->max_bind_tries)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }
        }
    }

    /* set connect timeout.. iPlanet SDK at this time */
#if defined(LDAP_X_OPT_CONNECT_TIMEOUT)
        timeout=cr->ldap_connect_timeout*1000;
        m_rerr(r,cr->debug,
                "[mod_auth_ldap.c (%d)] - Setting connect timeout to: %d seconds",MML,cr->ldap_connect_timeout);

        rc=ldap_set_option(cr->ld,LDAP_X_OPT_CONNECT_TIMEOUT,
                &timeout);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,cr->debug,
                "[mod_auth_ldap.c (%d)] - Could not set LDAP_X_OPT_CONNECT_TIMEOUT (%s)",MML,ldap_err2string(rc));
                                
        }
        else
        {
            m_rerr(r,cr->debug,
                "[mod_auth_ldap.c (%d)] - Successfully set connection timeout to %d seconds",MML,cr->ldap_connect_timeout);
        }
#else
        m_rerr(r,cr->debug,
                "[mod_auth_ldap.c (%d)] - you didn't compile with iPlanet C SDK, connect timeout will not be available",MML);
#endif /* LDAP_X_OPT_CONNECT_TIMEOUT */

    /*
    ** so far openldap sdk and server supports STARTTLS to encrypt
    ** traffic. muquit@muquit.com Dec-24-2002 
    */
#ifdef HAVE_LDAP_START_TLS_S
    if (cr->auth_ldapstarttls == 1)
    {
        ldap_vers=LDAP_VERSION3;
        rc=ldap_set_option(cr->ld,LDAP_OPT_PROTOCOL_VERSION,&ldap_vers);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,1,"[mod_auth_ldap.c] - ldap_set_option(LDAP_VERSION3) failed (%d): %s",rc,ldap_err2string(rc));
        }

        rc=ldap_start_tls_s(cr->ld,NULL,NULL);
        if (rc != LDAP_SUCCESS)
        {
            m_rerr(r,1,"[mod_auth_ldap.c] - Starting TLS failed (%d): %s",rc,ldap_err2string(rc));
            if (bind_try >= cr->max_bind_tries)
            {
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }
        }
    }
#endif /* HAVE_LDAP_START_TLS_S */


    if (!(cr->auth_on_bind))
    {
      /* now get the User DN */
      dn=mod_auth_ldap_find_user_dn(r,&try_again);

      if (dn == (char *) NULL)
      {
          m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - Bind attempt# %d, cound not find DN for user \"%s\" with attr \"%s\"",MML,bind_try,r->user,cr->uid_attr);

          /*
          ** It's possible already open connection by the module is closed 
          ** by the ldap server or ldap server is down. Some people 
          ** reported this kind of problem.  muquit@lucent.com Jun-09-2003 
          */
          if (try_again && (bind_try < cr->max_bind_tries))
          {
              ldap_unbind(cr->ld); /* we don't care if it fails */
              cr->ld=NULL;
              conns->ld=NULL;
              bind_try++;
              goto try_again;
          }
#if 0
          if (!try_again)
          {
              m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - Connection to LDAP server is open but user does not exist, will not retry",MML);
          }
#endif /* if 0 */

          /* pass control to lower modules if AuthLDAPAuthoritative=no */
          if (! (cr->auth_ldapauthoritative))
          {
              m_rerr(r,cr->debug,"[mod_auth_ldap.c] - Returning control to lower level modules");

              rcode=DECLINED;
              goto exit_processing;
          }
          if (try_again)
          {
            m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - Tried to bind %d times. Giving up.",
                    MML,bind_try);
          }
          rcode=HTTP_UNAUTHORIZED;
          goto exit_processing;
      }
    }
    else
    {
        m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - AuthOnBind is On, skipping LDAP Search",
                MML);

        /* Build our dn from Base_DN and the authenticating username */
        if (cr->sub_dn)
        {
            ap_snprintf(authonbind_dn,sizeof(authonbind_dn)-1,"%s=%s,%s,%s",
                cr->uid_attr,r->user,cr->sub_dn,cr->base_dn);
        }
        else
        {
            ap_snprintf(authonbind_dn,sizeof(authonbind_dn)-1,"%s=%s,%s",
                cr->uid_attr,r->user,cr->base_dn);
        }
        dn=authonbind_dn;
        m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - AuthOnBind is On, Using explicit user DN \"%s\"",
            MML,dn);
    }

    /* store user DN, we'll use it later if needed*/
    cr->user_dn=ap_pstrdup(r->pool,dn);

    m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - Bind attempt# %d, pid=%d,User DN: %s",
            MML,bind_try,getpid(),cr->user_dn);

    /*
    ** now if we can successfully bind to LDAP server as the specified
    ** userid and passoword, the user is a valid one!
    */
    if (ldap_simple_bind_s(cr->ld,dn,sent_pw) != LDAP_SUCCESS)
    {
        if (r->user)
            m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - Invalid Login or Password for: %s",MML,r->user);

        if (!(cr->auth_on_bind))
        {
            /* use found, probably failed due to wrong password , don't 
             * bother to try again
             */
            if (cr->user_dn)
            {
                if (! (cr->auth_ldapauthoritative))
                {
                    m_rerr(r,cr->debug,"[mod_auth_ldap.c] - Returning control to lower level modules");

                    rcode=DECLINED;
                    goto exit_processing;
                }
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }
        }

        /*
         * we didn't do any search so we don't if it failed because binding failed
         * or because the user does not exist
         */
        if (bind_try < cr->max_bind_tries)
        {
            ldap_unbind(cr->ld); /* we don't care if it fails */
            cr->ld=NULL;
            conns->ld=NULL;
            bind_try++;
            goto try_again;
        }

        m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - Tried to bind %d times. Giving up.", MML,bind_try);

        /* pass control to lower modules if AuthLDAPAuthoritative=no */
        if (! (cr->auth_ldapauthoritative))
        {
            m_rerr(r,cr->debug,"[mod_auth_ldap.c] - Returning control to lower level modules");

            rcode=DECLINED;
            goto exit_processing;
        }

        rcode=HTTP_UNAUTHORIZED;
        goto exit_processing;
    }

exit_processing:
    /* free things we won't need */
    if (dn != (char *) NULL)
    {
      if (!(cr->auth_on_bind))
      {
#if LDAP_SDK_VERSION == 2
        (void) free(dn);
#else
        ldap_memfree(dn);
#endif /* LDAP_SDK_VERSION != 2 */
      }
      dn=NULL;
    }

#if APR_HAS_THREADS
    /* Don't forget to release lock before returing*/
    if (cr->mutex)
    {
        apr_thread_mutex_unlock(cr->mutex);
    }
#endif /* APR_HAS_THREADS */

    if (rcode == OK)
    {
        /* Set environment variables */
        ap_table_setn(r->subprocess_env,"LDAP_USER",r->user);
        ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_VERSION",
            MOD_AUTH_LDAP_VERSION_S);

        /*
         * if UID_Attr_Use is set, search this attribute and set the
         * value to LDAP_USER_ALT env variable
         */
        if (cr->uid_attr_alt)
        {
            mod_auth_ldap_set_ldap_user_alt(r);
        }
        cr->env_vars_set=1;
    }

    /* If we've turned off persistent connections, close ours. */
    if (!(cr->ldap_persistent))
    {
      m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - LDAP_Persistent is off.  Closing old connection, pid=%d",MML,getpid());
      ldap_unbind(conns->ld);
      conns->ld = NULL;
    }

    if (rcode == HTTP_UNAUTHORIZED)
        ap_note_basic_auth_failure(r);

    return (rcode);
}

static int compare_group_recursively(request_rec *r,char *master_gr_dn)
{
    char
        *gr_dn=NULL;

    LDAPMessage
        *res=NULL,
        *ent=NULL;

    int
        retval=(-1),
        rc;

    ldap_auth_config_rec
        *cr;

    cr=(ldap_auth_config_rec *)
       ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    gr_dn=NULL;

    for(;;)
    {
        if (gr_dn)
        {
            mod_auth_ldap_free_dn(gr_dn);
            gr_dn=NULL;
        }

        if (res != NULL)
        {
            ldap_msgfree(res);
            res=NULL;
        }
        if (cr->debug)
        {
            m_rerr(r,cr->debug, "[mod_auth_ldap.c (%d)] - compare_group_recursively(), filter=%s",MML,cr->filter);
        }

        rc=ldap_search_s(cr->ld,
                cr->base_dn,
                LDAP_SCOPE_SUBTREE,
                cr->filter,
                NULL,
                0,
                &res);
        if (rc == LDAP_SUCCESS)
        {
            for (ent=ldap_first_entry(cr->ld,res); ent ;
                    ent=ldap_next_entry(cr->ld,ent))
            {
                gr_dn=ldap_get_dn(cr->ld,ent);
                if (gr_dn == NULL)
                {
                    retval=(-1);
                    goto exit_processing;
                }

                if (cr->debug)
                {
                    m_rerr(r,cr->debug,
                        "[mod_auth_ldap.c (%d)] - >>Trying Group: %s",MML,gr_dn);
                }

                rc=ldap_compare_s(cr->ld,
                        master_gr_dn,
                        cr->group_attr,
                        gr_dn);

                if (rc == LDAP_COMPARE_TRUE)
                {
                    m_rerr(r,cr->debug,
                        "[mod_auth_ldap.c (%d)] - ** Group matched: %s",MML,gr_dn);
                    mod_auth_ldap_free_dn(gr_dn);
                    gr_dn=NULL;

                    retval=(0);
                    goto exit_processing;
                }

                if (cr->debug)
                {
                    m_rerr(r,cr->debug,
                        "[mod_auth_ldap.c (%d)] - No match",MML);
                }

                ap_snprintf(cr->filter,sizeof(cr->filter)-1,
                    "(%s=%s)",cr->group_attr,gr_dn);
                res=NULL;
            }
            if (gr_dn == NULL)
            {
                retval=(-1);
                goto exit_processing;
            }
        }
        else
        {
            retval=(-1);
            goto exit_processing;
        }
    }

exit_processing:

    if (gr_dn)
    {
        mod_auth_ldap_free_dn(gr_dn);
        gr_dn=NULL;
    }

    if (res != NULL)
    {
        ldap_msgfree(res);
        res=NULL;
    }

    return(retval);
}


/**
 * return 0 if the user group belong to a master group
 *        -1 otherwise
 */
static int check_nested_groups(char *master_gr_dn,request_rec *r)
{
    ldap_auth_config_rec
        *cr;

    int
        gr_count=0,
        count,
        retval=(-1),
        rc;

    char
        *dn=NULL;

    LDAPMessage
        *res=NULL,
        *ent=NULL;

    cr=(ldap_auth_config_rec *)
       ap_get_module_config(r->per_dir_config,&auth_ldap_module);

    /* find all the groups the user belongs to */
    ap_snprintf(cr->filter,sizeof(cr->filter)-1,
            "(%s=%s)",cr->group_attr,cr->user_dn);
    rc=ldap_search_s(cr->ld,
            cr->base_dn,
            LDAP_SCOPE_SUBTREE,
            cr->filter,
            NULL,
            0,
            &res);
    if (rc != LDAP_SUCCESS)
    {
        if (cr->debug)
        {
            m_rerr(r,cr->debug,
              "[mod_auth_ldap.c (%d)] - check_nested_groups() user \"%s\" does not belong to any group",MML,r->user);
        }
        retval=(-1);
        goto exit_processing;
    }
    count=ldap_count_entries(cr->ld,res);
    if (count <= 0)
    {
        if (cr->debug)
        {
            m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - user \"%s\" does not belong to any group",MML,r->user);
        }
        retval=(-1);
        goto exit_processing;
    }

    if (cr->debug)
    {
        m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - user \"%s\" belongs to %d groups",MML,r->user,count);
    }

    /*
    ** now go through each group and see if the group itself or if the group
    ** belong to any other group and that group belong some other group and 
    ** so on eventually belong to the master group
    */
    for (ent=ldap_first_entry(cr->ld,res); ent != NULL;
            ent=ldap_next_entry(cr->ld,ent))
    {
        gr_count++;
        dn=ldap_get_dn(cr->ld,ent);
        if (dn)
        {
            if (cr->debug)
            {
                m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - >Trying (%d) Group: %s",MML,gr_count,dn);
            }

            /* check if this group belong to the master group */
            rc=ldap_compare_s(cr->ld,
                    master_gr_dn,
                    cr->group_attr,
                    dn);

            if (rc == LDAP_COMPARE_TRUE)
            {
                if (cr->debug)
                {
                    m_rerr(r,cr->debug,
                        "[mod_auth_ldap.c (%d)] - ** Group (%d) matched: %s",MML,gr_count,dn);
                }
                mod_auth_ldap_free_dn(dn);
                dn=NULL;

                retval=(0);
                goto exit_processing;
            }
            else
            {
                if (cr->debug)
                {
                    m_rerr(r,cr->debug,
                        "[mod_auth_ldap.c (%d)] - No match (%d)",MML,gr_count);
                }
            }

            ap_snprintf(cr->filter,sizeof(cr->filter)-1,
                    "(%s=%s)",cr->group_attr,dn);

            mod_auth_ldap_free_dn(dn);
            dn=NULL;

            /*
            ** the sub group does not belong to the master group. check
            ** if the subgroup belongs to other groups which may belong to
            ** the master group
            */
            rc=compare_group_recursively(r,master_gr_dn);
            if (rc == 0)
            {
                retval=0;
                goto exit_processing;
            }
        }
    }

exit_processing:

    if (res != NULL)
        ldap_msgfree(res);

    return(retval);
}

/* called to check to see if resource being requested requires authorization. */
static int check_user_authorization(request_rec *r)
{
    const apr_array_header_t
        *reqs_arr;

    require_line
        *reqs;

    register int
        x;

    int
        rcode=OK,
        rc,
        m;

    const char
        *t,
        *w;

    ldap_auth_config_rec
        *cr;

    conn_rec
        *c;

    char
        *attr,
        *user=r->user;

    cr=(ldap_auth_config_rec *) NULL;
    reqs_arr=(ap_array_header_t *)NULL;
    reqs=(require_line *) NULL;

    cr=(ldap_auth_config_rec *)
       ap_get_module_config(r->per_dir_config,&auth_ldap_module);
    reqs_arr=ap_requires(r);
    reqs=reqs_arr ? (require_line *) reqs_arr->elts : NULL;
    m=r->method_number;
    c=r->connection;

    /* mm, Aug-20-1999 */
    if (cr->ldap_server == NULL)
        return(DECLINED);

    if (cr->ld == (LDAP *) NULL)
    {
        m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - no open LDAP connection",
                MML);
        return (DECLINED);
    }

#if APR_HAS_THREADS
    /* lock before using cr->ld*/
    if (cr->mutex)
    {
        apr_thread_mutex_lock(cr->mutex);
    }
#endif /* APR_HAS_THREADS */

    /* server cfg file must have at least one require line */
    if (reqs_arr == NULL)
    {
        m_rerr(r,1,"[mod_auth_ldap.c] - no require line in config file");

        rcode=DECLINED;
        goto exit_processing;
    }

    for (x=0; x < reqs_arr->nelts; x++)
    {
        if (!(reqs[x].method_mask & (1 << m)))
            continue;
        t=reqs[x].requirement;

        /*
        ** use ap_getword_white() instead of ap_getword() and my own rmallws
        ** muquit@muquit.com, Feb-10-2000
        */
        w=ap_getword_white(r->pool,&t);

        /*
        ** if "require valid-user" found in cfg, the user is authorized.
        ** we'r outta here with OK
        */
        if(strcmp(w,"valid-user") == 0)
        {
            /* remove later muquit@muquit.com */
            if (cr->user_dn != NULL)
            {
                m_rerr(r,cr->debug,"mod_auth_ldap.c (%d) ] - user dn obtained earlier: %s",MML,cr->user_dn);
            }

            rcode=OK;
            goto exit_processing;
        }
        else if (strcmp(w,"user") == 0)
        {
            /*
            ** allow in only if the userid is one in the space separated list
            ** in the server cfg file. now t should have all the space 
            **  separated uids from cfg file 
            */
            while (*t != '\0')
            {
                /*
                ** go through all the uids and compare them with the uid
                ** from connection (which is in the LDAP server)
                */
                w=ap_getword_conf(r->pool,&t);


                m_rerr(r,cr->debug,"[mod_auth_ldap.c] - w=\"%s\"",w);
                if (*w != '\0')
                {
                    if (strcmp(w,user) == 0)
                    {
                        m_rerr(r,cr->debug,
                            "[mod_auth_ldap.c (%d)] - user %s is authorized",MML,user);
                        rcode=OK;
                        goto exit_processing;
                    }
                }
            }
            rcode=HTTP_UNAUTHORIZED;
            goto exit_processing;
        }
        else if (strcmp(w,"filter") == 0)
        {
            int
                rc;

            LDAPMessage
                *result=NULL;
            char
                *filter=NULL;

            /*
            ** We'll check if the user match the attributes in filter.
            ** We're going to use the user_dn obtained earlier, and we'll
            ** use the same open connection.
            */
            if (cr->ld == (LDAP *) NULL)
            {
                m_rerr(r,1,"[mod_auth_ldap.c (%d)] - check_user_authorization() no open LDAP connection",MML);
                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }

            if (cr->user_dn == NULL)
            {
                m_rerr(r,1,"[mod_auth_ldap.c (%d)] - check_user_authorization() no user DN available",MML);

                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }

            filter=ap_getword_conf(r->pool,&t);
            rc=ldap_search_s(cr->ld,
                             cr->user_dn,
                             LDAP_SCOPE_SUBTREE,
                             filter,
                             NULL,
                             0,
                             &result);
            if (ldap_first_entry (cr->ld,result) == NULL)
            {
                if (result != NULL)
                    ldap_msgfree(result);

                m_rerr(r,cr->debug,"[mod_auth_ldap.c] - check_user_authorization() filter %s did not match  for user %s",filter,user);

                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }

            if (rc == LDAP_SUCCESS && result != NULL)
            {
                m_rerr(r,cr->debug,"[mod_auth_ldap.c] - check_user_authorization() filter %s matched for user %s",filter,user);

#ifdef MY_DEBUG /* for me only */
                dumpResult(cr->ld,result,r);
#endif /* MY_DEBUG */

                if (result != NULL)
                    ldap_msgfree(result);

                rcode=OK;
                goto exit_processing;
            }

            if (result != NULL)
                ldap_msgfree(result);

            m_rerr(r,cr->debug,
                          "[mod_auth_ldap.c] - check_user_authorization() filter %s did not match  for user %s",filter,user);

            rcode=HTTP_UNAUTHORIZED;
            goto exit_processing;
        }
        else if (strcmp(w,"group") == 0)
        {
            char
                *group_attrib=NULL,
                *group_name=NULL,
                szgdn[1024];

            group_attrib=ap_pstrdup(r->pool,w);
            rmallws(group_attrib);

            m_rerr(r,cr->debug,
                          "[mod_auth_ldap.c (%d)] - ldap attribute=%s",
                          MML,group_attrib);
            /*
            ** XXX why not using the dn we found in user access??
            ** muquit@muquit.com 
            */
            if (cr->user_dn == (char *) NULL)
            {
                m_rerr(r,cr->debug,
                   "[mod_auth_ldap.c] -  no dn found for user in group attrib");

                /* pass control to lower modules if AuthLDAPAuthoritative=no */
                if (! (cr->auth_ldapauthoritative))
                {
                    m_rerr(r,cr->debug,
                    "[mod_auth_ldap.c] - returning control to lower modules");
                    rcode=DECLINED;
                    goto exit_processing;
                }
                else
                {
                    m_rerr(r,cr->debug,"[mod_auth_ldap.c] - AuthLDAPAuthoritative=on");
                }

                rcode=HTTP_UNAUTHORIZED;
                goto exit_processing;
            }

            while (*t != '\0')
            {
                /*
                ** use ap_getword_conf() instead of the above. ap_getword()
                ** tokenizes the string, so a group like "foo group" can not
                ** be used. -- muquit@muquit.com, Feb-10-2000
                */
                w=ap_getword_conf(r->pool,&t);
                if (*w != '\0')
                {
                    if (cr->group_attr)
                    {
                        m_rerr(r,cr->debug,
                            "[mod_auth_ldap.c (%d)] - Group_Attr=%s",
                            MML,cr->group_attr);
                    }
                    ap_snprintf(szgdn,sizeof(szgdn)-1,"%s,%s",w,cr->base_dn);
                    m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - Group DN:%s",
                            MML,szgdn);
                    rc=ldap_compare_s(cr->ld,
                                      szgdn,
                                      cr->group_attr, /*uniqueMember */
                                      cr->user_dn);

                    if (rc == LDAP_COMPARE_TRUE)
                    {
                        m_rerr(r,cr->debug,
                             "[mod_auth_ldap.c (%d)] - user \"%s\" belongs to group \"%s\" directly",MML,r->user,w);

                        rcode=OK;
                        goto exit_processing;
                    }
                    else if (rc == LDAP_COMPARE_FALSE)
                    {
                        m_rerr(r,cr->debug,
                                "[mod_auth_ldap.c] %d - user \"%s\" does not belong to group \"%s\" directly",MML,r->user,w);

                        if (cr->support_nested_groups)
                        {
                            m_rerr(r,cr->debug,"[mod_auth_ldap.c] %d - Checking nested groups",MML);
                            rc=check_nested_groups(szgdn,r);
                            if (rc == 0) /* ok */
                            {
                                m_rerr(r,cr->debug,
                                    "[mod_auth_ldap.c] %d - user \"%s\" belongs to group \"%s\" by nested groups",MML,r->user,w);

                                rcode=OK;
                                goto exit_processing;
                            }
                        }
                    }
                    else
                    {
#ifdef DEBUG_LDAP
                        ldap_perror(cr->ld,"[mod_auth_ldap.c] Error:");
#endif /* DEBUG_LDAP */
                    }
                }
            }   /* while */

            if (! (cr->auth_ldapauthoritative))
            {
                m_rerr(r,cr->debug,"[mod_auth_ldap.c] - returning control to lower modules");
                rcode=DECLINED;
                goto exit_processing;
            }
            rcode=HTTP_UNAUTHORIZED;
            goto exit_processing;
        }
        else
        {
            /*
            ** else allow in only if the ldap attribute is present in the LDAP
            ** server for the user
            */
            if (*w != '\0')
                attr=ap_pstrdup(r->pool,w);
            m_rerr(r,cr->debug,"[mod_auth_ldap.c] - ldap attribute=%s",w);

            /* go through the attribute values and check if the user has it */
            while (*t != '\0')
            {
                /*
                ** use ap_getword_conf() instead of the above. ap_getword()
                ** tokenizes string, so a attribute like "Foo org" can not
                ** be used.  -- muquit@muquit.com, Feb-10-2000
                */
                w=ap_getword_conf(r->pool,&t);
                if (*w != '\0')
                {
                    rc=ldap_compare_s(cr->ld,cr->user_dn,attr,w);
                    if (rc == LDAP_COMPARE_TRUE)
                    {
                        m_rerr(r,cr->debug,"[mod_auth_ldap.c] - user \"%s\" has attribute %s=\"%s\"",user,attr,w);

                        rcode=OK;
                        goto exit_processing;
                    }
                    else
                    {
                        m_rerr(r,cr->debug,"[mod_auth_ldap.c] %d - user %s does not have attribute  %s=%s",MML,user,attr,w);
                    }
                }
            }
            rcode=HTTP_UNAUTHORIZED;
            goto exit_processing;
        }
    }

exit_processing:

    /*
     * XXX crashes windows under heavy load in ssl. it crashes in 
     * ldap_simple_bind_s() routine. why?
     * so for now I'll commentit out for windows.. might cause handle leak!
     * muquit@muquit.com Dec-31-2002 
     */
    if (cr->ld != (LDAP *) NULL)
    {
#if defined(WIN32)
        if (cr->start_ssl)
            goto there;
#endif /* WIN32 */

        /*
        if (ldap_unbind(cr->ld) == LDAP_SUCCESS)
        {
            m_rerr(r,cr->debug,"[mod_auth_ldap.c] %d - closed connection to LDAP server: %s, port:%d",MML,cr->ldap_server,cr->ldap_port);
            cr->ld=NULL;
        }
        */
        /*
        m_rerr(r,cr->debug,"[mod_auth_ldap.c] %d - closing connection to LDAP server: %s, port:%d",MML,cr->ldap_server,cr->ldap_port);
        ldap_unbind(cr->ld);
        cr->ld=NULL;
        */
    }
there:

    if (rcode != OK)
    {
        if (user)
            m_rerr(r,cr->debug,"[mod_auth_ldap.c (%d)] - user \"%s\" is not allowed",MML,user);
    }

#if APR_HAS_THREADS
    /* Don't forget to release lock before returing */
    if (cr->mutex)
    {
        apr_thread_mutex_unlock(cr->mutex);
    }
#endif /* APR_HAS_THREADS */
    if (rcode == OK)
    {
        if (!cr->env_vars_set)
        {
            /* Set LDAP_USER environment variables in all cases */
            ap_table_setn(r->subprocess_env,"LDAP_USER",r->user);
            ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_VERSION",
                MOD_AUTH_LDAP_VERSION_S);

            if (cr->uid_attr_alt)
            {
                mod_auth_ldap_set_ldap_user_alt(r);
            }
        }
    }
    if (rcode == HTTP_UNAUTHORIZED)
        ap_note_basic_auth_failure(r);

    return (rcode);
}

/* create per-server configuration */
void *create_server_config(apr_pool_t *p,server_rec *s)
{
    Conns *conns= (Conns *) apr_palloc(p,sizeof(Conns));

    if (conns == NULL)
        return(NULL);

    conns->ld=NULL;
    return(conns);
}

static void register_hooks(apr_pool_t *p)
{
    ap_hook_check_user_id(check_user_authentication,
                          NULL,NULL,
                          APR_HOOK_MIDDLE);

    ap_hook_auth_checker(check_user_authorization,
                         NULL,NULL,
                         APR_HOOK_MIDDLE);
}

/*--------------------------------------------------------------------------*/
/* Finally, the list of callback routines and data structures that          */
/* provide the hooks into our module from the other parts of the server.    */
/*--------------------------------------------------------------------------*/
module AP_MODULE_DECLARE_DATA auth_ldap_module =
{
    STANDARD20_MODULE_STUFF,
    create_ldap_auth_dir_config,  /* per-directory config creator */
    NULL,                         /* dir config merger */
    create_server_config,         /* server config creator */
    NULL,                         /* server config merger */
    ldap_auth_cmds,               /* command table */
    register_hooks                /* register hooks */
};
